

def test_create_stat_table():
    pass


def test_get_latest_successful_ts():
    pass


def test_update_latest_successful_ts():
    pass


def test_create_oplog_table():
    pass


def test_log_rows():
    pass
